﻿/* ETBdocNav.jsx ====== UTF-8 =====================================================================
   Display a tree view from paragraphs for document navigation
Location      E:\_DDDprojects\FM-toolbar00\AllETB\Scripts\StandAloneVersions\ETBdocNav.jsx 
Documents     www.daube.ch/docu/fmaker84.html
Attention     The structural levels are taken from the paragraph tag names by regular expression
              or table in ETBextras.cfg
Reference     Rick Quatro: DocumentNavigator.jsx

Copyleft      2020 Klaus Daube, Zürich, Switzerland, klaus@daube.ch
History       2020-01-24 start of development
              2020-03-07 stand alone version for fmaker84.html
*/ ; // ===========================================================================================

#target framemaker

if (typeof KLD_E == "undefined") {
  KLD_E = {};                                     // Global object may already be defined by others
}
KLD_E.sExtrasCfg = "ETBextras.cfg";               // If the reading routines would also be in sub-dir
KLD_E.oLevelSettings = {};
KLD_E.oTree = {};
KLD_E.sDnav_title     = { en: "Document Navigation ETB",
                        de: "Dokument Navigation ETB",
                        fr: "Navigation de document ETB"};
KLD_E.sDnav_docRequ   = { en: "An acitve document is required for this function.",
                        de: "Ein aktives Dokument ist notwendig für diese Funktion.",
                        fr: "Un document actif est nécessaire pour cette fonction."};
KLD_E.sDnav_lvlDefs   = { en: "No (correct) table defined to evaluate the level of headings",
                        de: "Keine (gültige) Tabelle ist definiert zur Evaluation der Überschriften-Ebenen",
                        fr: "Pas de tableau (correct) défini pour évaluer des niveaux des titres"};
KLD_E.sDnav_noPgfTab  = { en: "No paragraphs found according to the level-table",
                        de: "Keine Absätze gefunden, die der Ebenen-Definition in der Tabelle entsprechen.",
                        fr: "Aucun paragraphe trouvé ne correspond à la définition du niveau dans le tableau."};
KLD_E.sDnav_hlpTip1   = { en: "Go back",
                        de: "Zurück",
                        fr: "Revenir en arrière"};
KLD_E.sDnav_bExpand   = { en: "Expand",
                        de: "Erweitern",
                        fr: "Développer"};
KLD_E.sDnav_hlpTip2   = { en: "Expand all nodes",
                        de: "Alle Zweige erweitern",
                        fr: "Développer tous les nœuds"};
KLD_E.sDnav_bCollapse = { en: "Collapse",
                        de: "Reduzieren",
                        fr: "Réduire"};
KLD_E.sDnav_hlpTip3   = { en: "Collapse all branches",
                        de: "Alle Zweige reduzieren",
                        fr: "Réduire toutes les branches"};
KLD_E.sDnav_hlpTip4   = { en: "Go back to previous target",
                        de: "Zu früherem Ziel zurück gehen",
                        fr: "Revenir à l'objectif précédent"};
KLD_E.sDnav_cbStyle   = { en: "Style name / Title",
                        de: "Formatname / Titel",
                        fr: "Nom du style / Titre"};
KLD_E.sDnav_hlpTip5   = { en: "Set this option to display Tag names rather than Title text → then Refresh",
                        de: "Option aktivieren, um Formatnamen statt der Titel anzuzeigen → dann Aktualisieren",
                        fr: "Réglez cette option pour afficher les noms des balises plutôt que le texte du titre → puis Rafraîchir"};
KLD_E.sDnav_bRefresh  = { en: "Refresh",
                        de: "Aktualisieren",
                        fr: "Actualiser"};
KLD_E.sDnav_hlpTip6   = { en: "Refresh the palette",
                        de: "Titel-Liste aktualisieren",
                        fr: "Actualiser la liste des titres"};
KLD_E.sDnav_missing   = { en: "→ Missing level",
                        de: "→ Fehlende Ebene",
                        fr: "→ Niveau manquant"};
KLD_E.sDnav_unused    = { en: "→ Unused level",
                        de: "→ Nicht verwendete Ebene",
                        fr: "→ Niveau non utilisé"};
KLD_E.sDnav_bDefine   = { en: "Define",
                        de: "Definieren",
                        fr: "Définir"};
KLD_E.sDnav_hlpTip7   = { en: "Define the heading levels",
                        de: "Überschriften-Ebenen definieren",
                        fr: "Définir les niveaux des titres"};
KLD_E.sDnav_tLevel    = { en: "Level",
                        de: "Ebene",
                        fr: "Niveau"};
KLD_E.sDnav_tExplain  = { en: "Format names, separated by semicolon. No space after the semicolon.",
                        de: "Format-Namen durch Strichpunkt getrennt. Keine Leerstelle nach dem Strichpunkt.",
                        fr: "Noms de format, séparés par un point-virgule. Pas d'espace après le point-virgule."};
KLD_E.sDnav_btnSave   = { en: "Save",
                        de: "Speichern",
                        fr: "Enregistrer"};
KLD_E.sDnav_btnCancel = { en: "Cancel",
                        de: "Abbrechen",
                        fr: "Annuller"};

KLD_E.DefineLevels = function () {
var wDefine, sLevel0, sLevel1, sLevel2, sLevel3;

  wDefine = new Window ("dialog", localize (KLD_E.sDnav_hlpTip7), undefined); // localize (KLD_E.sDnav_title)
  wDefine.preferredSize.width = 500; 
  wDefine.alignChildren = ["left","top"]; 
  wDefine.spacing = 10; 
  wDefine.margins = 5; 
  wDefine.p1 = wDefine.add("panel", undefined, undefined); 
  wDefine.p1.orientation = "column"; 
  wDefine.p1.alignChildren = ["left","top"]; 

  wDefine.p1.gt = wDefine.p1.add("group", undefined, undefined); 
  wDefine.p1.gt.stx = wDefine.p1.gt.add("statictext", undefined, localize (KLD_E.sDnav_tLevel)); 
  wDefine.p1.gt.sty = wDefine.p1.gt.add("statictext", undefined, localize (KLD_E.sDnav_tExplain)); 

  wDefine.p1.g0 = wDefine.p1.add("group", undefined, undefined); 
  wDefine.p1.g0.st0 = wDefine.p1.g0.add("statictext", undefined, "0"); 
  wDefine.p1.g0.e0 = wDefine.p1.g0.add("edittext" , undefined, "level_0-format"); 
  wDefine.p1.g0.e0.graphics.font = ScriptUI.newFont ("Consolas", "", 11);
  wDefine.p1.g0.e0.preferredSize.width = 450; 

  wDefine.p1.g1 = wDefine.p1.add("group", undefined, undefined); 
  wDefine.p1.g1.st0 = wDefine.p1.g1.add("statictext", undefined, "1"); 
  wDefine.p1.g1.e0 = wDefine.p1.g1.add("edittext" , undefined, "level_1-format"); 
  wDefine.p1.g1.e0.graphics.font = ScriptUI.newFont ("Consolas", "", 11);
  wDefine.p1.g1.e0.preferredSize.width = 450; 

  wDefine.p1.g2 = wDefine.p1.add("group", undefined, undefined); 
  wDefine.p1.g2.st0 = wDefine.p1.g2.add("statictext", undefined, "2"); 
  wDefine.p1.g2.e0 = wDefine.p1.g2.add("edittext" , undefined, "level_2-format"); 
  wDefine.p1.g2.e0.graphics.font = ScriptUI.newFont ("Consolas", "", 11);
  wDefine.p1.g2.e0.preferredSize.width = 450; 

  wDefine.p1.g3 = wDefine.p1.add("group", undefined, undefined); 
  wDefine.p1.g3.st0 = wDefine.p1.g3.add("statictext", undefined, "3"); 
  wDefine.p1.g3.e0 = wDefine.p1.g3.add("edittext" , undefined, "level_3-format"); 
  wDefine.p1.g3.e0.graphics.font = ScriptUI.newFont ("Consolas", "", 11);
  wDefine.p1.g3.e0.preferredSize.width = 450; 
// --- button group ------------------------------
  wDefine.p1.gb = wDefine.p1.add("group", undefined, undefined); 
  wDefine.p1.gb.alignChildren = ["left","top"]; 
  wDefine.p1.gb.btnSave = wDefine.p1.gb.add("button", undefined, localize (KLD_E.sDnav_btnSave)); 
  wDefine.p1.gb.spacer  = wDefine.p1.gb.add("statictext", undefined, "   ");
  wDefine.p1.gb.spacer.preferredSize.width = 285;
  wDefine.p1.gb.btnCancel = wDefine.p1.gb.add("button", undefined, localize (KLD_E.sDnav_btnCancel)); 
// --- behaviour of the edit fields --------------
  sLevel0 = KLD_E.GetIniValue (KLD_E.sExtrasCfg, "0", "level-0-formats");
  wDefine.p1.g0.e0.text = sLevel0;
  sLevel1 = KLD_E.GetIniValue (KLD_E.sExtrasCfg, "1", "level-1-formats");
  wDefine.p1.g1.e0.text = sLevel1;
  sLevel2 = KLD_E.GetIniValue (KLD_E.sExtrasCfg, "2", "level-2-formats");
  wDefine.p1.g2.e0.text = sLevel2;
  sLevel3 = KLD_E.GetIniValue (KLD_E.sExtrasCfg, "3", "level-3-formats");
  wDefine.p1.g3.e0.text = sLevel3;
// --- behaviour of the buttons ------------------
  wDefine.p1.gb.btnSave.onClick = function () {
    if (sLevel0 != wDefine.p1.g0.e0.text) {
      KLD_E.PutIniValue (KLD_E.sExtrasCfg, "0", wDefine.p1.g0.e0.text);
    }
    if (sLevel1 != wDefine.p1.g1.e0.text) {
      KLD_E.PutIniValue (KLD_E.sExtrasCfg, "1", wDefine.p1.g1.e0.text);
    }
    if (sLevel2 != wDefine.p1.g2.e0.text) {
      KLD_E.PutIniValue (KLD_E.sExtrasCfg, "2", wDefine.p1.g2.e0.text);
    }
    if (sLevel3 != wDefine.p1.g3.e0.text) {
      KLD_E.PutIniValue (KLD_E.sExtrasCfg, "3", wDefine.p1.g3.e0.text);
    }
    wDefine.close();
    return;
  }
  wDefine.p1.gb.btnCancel.onClick = function () {
    wDefine.close();
  }
// --- final tasks -------------------------------
  wDefine.show();  
} // --- end KLD_E.DefineLevels

KLD_E.DocNav  = function (location, bTag)  { // === Main function =================================
// Arguments  location       Current screen location of palette. used for resize
// Called by  -
// Calling    KLD_E.BuildPalette, KLD_E.GetParagraphs
// Comment    To be implemented in ETBdocNav

var aPgfs = [], oDoc, wDocNav;

  if (bTag === undefined) {bTag = false};
  oDoc = app.ActiveDoc;
  if (!oDoc.ObjectValid()) {
    Alert (localize (KLD_E.sDnav_docRequ), localize (KLD_E.sDnav_title), true);
    return;
  }
  KLD_E.GetLevelSettings ();
  KLD_E.GetParagraphs (oDoc, aPgfs, KLD_E.sFormats);
  if (aPgfs.length == 0) {
    alert (localize (KLD_E.sDnav_noPgfTab), localize (KLD_E.sDnav_title), true);
    KLD_E.DefineLevels ();                        // allow user to correct the table.
    return;
  };
  wDocNav = KLD_E.BuildPalette (oDoc, aPgfs, KLD_E.sFormats, bTag); 
  if (location) {
    wDocNav.bounds = location;  // Restore the palette's previous size and location on the screen.
  }
} // end of KLD_E.DocNav

KLD_E.AddNodeOrItem = function (oTree, indxOnLevel, lvl, bNode, sLabel, sDocName, nUnique) { // ===
//            Build the tree item (node or item) and add the properties for navigation
// Arguments  oTree          TreeView object to be handled
//            indxOnLevel    Array of current indes in level. E.g. [1.0.2.0] will refer to
//                           oTree.items[1].items[0].items[2].items[0]
//            lvl            Level of heading (0: topmost, 3 (or else) lowest
//            bNode          true for node, false for item
//            sLabel         Label for the presentation of the node/item
//            sDocName       document name for the hypertext command
//            nUnique        the unique number of the target paragraph
// Returns    Modified oTree
// Called by  KLD_E.BuildTree
// Reference  Derived from the static model in the Adobe script examples
//            and Rick Quatro DocumentNavigotor.jsx
var bNode, oTmp = oTree, indx, q;
// $.writeln(indxOnLevel + "\t" + lvl + " " + bNode + "\t" + nUnique + " " + sLabel);
  for (q = 0; q < lvl; q++) {                     // This must NOT be q <= lvl !
    indx = indxOnLevel[q];
    oTmp = oTmp.items[indx];
  }
  if (bNode) {
    oTmp.add("node", sLabel);
  } else {
    oTmp.add("item", sLabel);
  }
  if (sDocName != null) {                         // add properties for navigation
    indx = indxOnLevel[q];
    oTmp.items[indx].navName = sDocName;
    oTmp.items[indx].navUnique = nUnique;
  }
} // end of KLD_E.AddNodeOrItem

KLD_E.BtnCollaps = function (oTree) { // ===- Collapse 'all' nodes in tree ========================
// Arguments  oTree          TreeView object to be handled
// Returns    Modified oTree
// Called by  KLD_E.BuildPalette
var j, branches;
  oTree.expanded = false;
  branches = oTree.items;
  for (var j = 0; j < branches.length; j++) {
    if (branches[j].type == 'node') {
      KLD_E.BtnCollaps (branches[j]);
    }
  }
} // end of KLD_E.BtnCollaps


KLD_E.BtnDefine = function () { // === Define heading level in cfg file ===========================
// Returns    modified ETBextras.cfg file
// Called by  KLD_E.BuildPalette

  KLD_E.DefineLevels ();
} // end of KLD_E.BtnDefine

KLD_E.BtnExpand = function (oTree) { // === Expand tree - used recursively ========================
// Arguments  oTree          TreeView object to be handled
// Returns    Modified oTree
// Called by  KLD_E.BuildPalette
// Reference  Peter Kharel: ScriptUI for dummies https://creativepro.com/files/kahrel/indesign/scriptui.html
var j, branches;
  oTree.expanded = true;
  branches = oTree.items;
  for (var j = 0; j < branches.length; j++) {
    if (branches[j].type == 'node') {
      KLD_E.BtnExpand (branches[j]);
    }
  }
} // end of KLD_E.BtnExpand

KLD_E.BuildPalette = function (oDoc, aPgfs, sFormats, bTag) { // === Build the palette ============
// Arguments  oDoc           The document to be analysed and navigated - the current document
//            aPgfs          Array of Paragraph objects fulfilling the Title criteria
//            sFormats Level settings (table)
// Returns    Window wDocNav
// Called by  KLD_E.DocNav
// Calling    KLD_E.BuildTree, KLD_E.BtnCollaps, KLD_E.BtnExpand, 
// Reference  Lots of experiments
var lastSlash, sRootName, oSettings = {};

  sRootName = oDoc.Name;
  lastSlash = sRootName.lastIndexOf("\\");
  sRootName = sRootName.substring(lastSlash+1);
  wDocNav = new Window ("palette", localize (KLD_E.sDnav_title), undefined, {resizeable: true}); // Navigation de document ETB
  KLD_E.GetPalSettings (oSettings);
  wDocNav.location = [oSettings.nLeft, oSettings.nTop + 30];  // wintitle is 30 pix high
// --- top group: buttons ------------------------
  wDocNav.g1 = wDocNav.add ("group", undefined);
  wDocNav.g1.alignment = ["","top"];
  wDocNav.g1.btnExpand = wDocNav.g1.add ("button", undefined, localize (KLD_E.sDnav_bExpand)); // Développer
  wDocNav.g1.btnExpand.size = [60,20];
  wDocNav.g1.btnExpand.helpTip = localize (KLD_E.sDnav_hlpTip2);
  wDocNav.g1.btnCollapse = wDocNav.g1.add ("button", undefined, localize (KLD_E.sDnav_bCollapse)); // Réduire
  wDocNav.g1.btnCollapse.size = [60,20];
  wDocNav.g1.btnCollapse.helpTip = localize (KLD_E.sDnav_hlpTip3);
  wDocNav.g1.btnDefine = wDocNav.g1.add ("button", undefined, localize (KLD_E.sDnav_bDefine)); // Réduire
  wDocNav.g1.btnDefine.size = [60,20];
  wDocNav.g1.btnDefine.helpTip = localize (KLD_E.sDnav_hlpTip7);
// --- main panel: the gree view -----------------
  wDocNav.p1 = wDocNav.add("panel", undefined, sRootName); // Root required for treeview to work
  wDocNav.p1.alignment = ["fill", "fill"];
  wDocNav.p1.alignChildren = ["fill", "fill"]

  KLD_E.oTree = wDocNav.p1.add ("treeview", undefined, "");
  KLD_E.oTree.preferredSize = [oSettings.nTwidth, oSettings.nTheight];
  KLD_E.oTree.graphics.font = ScriptUI.newFont (KLD_E.oTree.graphics.font.family,KLD_E.oTree.graphics.font.style,14);
  KLD_E.BuildTree (oDoc, KLD_E.oTree, aPgfs, sFormats, bTag);
// --- bottom group: buttons ---------------------
  wDocNav.g2 = wDocNav.add ("group", undefined);
  wDocNav.g2.alignment =  ["", "bottom"];
  wDocNav.g2.alignChildren = "fill";
//wDocNav.g2.orientation =  ["row"]; // default for group
  wDocNav.g2.btnBack = wDocNav.g2.add ("button",undefined,"<<");      // Go back button.
  wDocNav.g2.btnBack.size = [30,20];
  wDocNav.g2.btnBack.helpTip = localize (KLD_E.sDnav_hlpTip4);
  wDocNav.g2.cb1 = wDocNav.g2.add ("checkbox", undefined, localize (KLD_E.sDnav_cbStyle)); // Style name/Title ...
  wDocNav.g2.cb1.helpTip = localize (KLD_E.sDnav_hlpTip5);
  wDocNav.g2.cb1.value = bTag;
  wDocNav.g2.btnRefresh = wDocNav.g2.add ("button",undefined,localize (KLD_E.sDnav_bRefresh)); // Refresh Aktualisieren Actualiser
  wDocNav.g2.btnRefresh.helpTip = localize (KLD_E.sDnav_hlpTip6);
// --- behaviour of window and controls ----------
  wDocNav.onResize = function () {
    this.layout.resize();
  }
  KLD_E.oTree.onChange = function () {
    KLD_E.Navigate (oDoc, this);
  }
  wDocNav.g1.btnExpand.onClick = function () {
    KLD_E.BtnExpand(KLD_E.oTree);
  }
  wDocNav.g1.btnCollapse.onClick = function () {
    KLD_E.BtnExpand(KLD_E.oTree);
    KLD_E.BtnCollaps(KLD_E.oTree);
    wDocNav.show();
  }
  wDocNav.g1.btnDefine.onClick = function () {
    KLD_E.BtnDefine ();
    wDocNav.show();
  }
  wDocNav.g2.cb1.onClick = function () {
    bTag = wDocNav.g2.cb1.value;
  }
  wDocNav.g2.btnBack.onClick = function () {
    oDoc.HypertextCommand ("previouslink");
  }
  wDocNav.g2.btnRefresh.onClick = function () {
    var location = this.parent.parent.bounds;
    this.parent.parent.close ();
    KLD_E.DocNav (location, bTag); 
  }
// --- final actions -----------------------------
  KLD_E.BtnExpand (KLD_E.oTree)
  wDocNav.show();
  return wDocNav;                                 // required for Refresh
} // end of KLD_E.BuildPalette

KLD_E.BuildTree = function (oDoc, oTree, aPgfs, sFormats, bTag) { // === Build the tree ===========
// Arguments  oDoc           The document to be analysed and navigated - the current document
//            oTree          TreeView object to be set up
//            aPgfs          Array of Paragraph objects fulfilling the Title criteria
//            sFormats       Table to find the title paragraphs
//            bTag           true for presentation of tag name, false for title text
// Returns    The set up KLD_E.oTree
// Called by  KLD_E.BuildPalette
// Calling    KLD_E.AddNodeOrItem
// Reference  Lots of experiments
var bNode, indxOnLvl=[0], lvl=0, lvlNxt, lvlOld=0, j, jMax, 
    sDocname, sLabel, sTagName;

  sDocName = oDoc.Name.replace (/\\/g, "/");      // to be used in navigate
  jMax = aPgfs.length;
// --- Is structure complete? 
  lvlNxt = KLD_E.GetLevel (aPgfs[0], sFormats);
  while (lvlNxt > lvl) {   
    sLabel = localize (KLD_E.sDnav_unused);             // artificially filled
    KLD_E.AddNodeOrItem (oTree, indxOnLvl, lvl, true, sLabel, null, null)
    indxOnLvl.push(0);
    lvl = Number(lvl) + 1;
  }
// --- Build tree from paragraph levels ----------
  for (j = 0; j < jMax; j ++) {                   // for all extracted paragraphs
    sTagName = aPgfs[j].Name;
    unique = aPgfs[j].Unique;
    if (bTag) {
      sLabel = sTagName;
    } else {
//    sLabel = CleanPgfNumber(aPgfs[j]) + KLD_E.GetText (aPgfs[j]); // don't know what this is good for.
      sLabel = KLD_E.GetText (aPgfs[j]);
    }
    lvlOld = lvl;
    lvl = KLD_E.GetLevel (aPgfs[j], sFormats);
    lvlNxt = j+1 < jMax ? KLD_E.GetLevel (aPgfs[j+1], sFormats) : undefined // last item undefined
    bNode = lvlNxt > lvl;
    while (lvl < lvlOld) {                      // ascend from deeper level
      lvlOld -= 1;
      indxOnLvl.pop();
      indxOnLvl[lvlOld] = indxOnLvl[lvlOld] + 1;
    }
    if (bNode) {
      KLD_E.AddNodeOrItem (oTree, indxOnLvl, lvl, bNode, sLabel, sDocName, unique)
      indxOnLvl.push(0);                          // sub items counter 
    } else {
      KLD_E.AddNodeOrItem (oTree, indxOnLvl, lvl, bNode, sLabel, sDocName, unique)
      indxOnLvl[lvl] = indxOnLvl[lvl] + 1;          // inc counter
    }
// --- check for irregular diving ----------------
    while (lvlNxt - lvl > 1) {
      sLabel = localize (KLD_E.sDnav_missing);           // artificially filled
      lvl = Number(lvl) + 1;
      KLD_E.AddNodeOrItem (oTree, indxOnLvl, lvl, true, sLabel, null, null)
      indxOnLvl.push(0);
    }
  }
} // end of KLD_E.BuildTree

KLD_E.GetIniValue = function (sCfgFile, sValueName, vDefault) { // === Read an ini item into a variable  <<<< approx in ETBinfo → move to ETBextras
// Arguments  sCfgFile       ini file name (e.g. "
//            sValueName     is that used in the ini-file (e.g. "FMhome")
//            vDefault       default value, if not found in file
// Returns    Value from the ini file (string)
// Attention  Windows ini files are still only CP 1252, whareas maker.ini can be UTF-8 also
// Comment    Empty lines, comments and section names are skipped
// Used in   FMbiblio, KLD_E.SetUpMenu
var regex, settingsFile, sFoundValue, thisLine, match, notFound = true, message; 
    regex = {
      section: /^\s*\[\s*([^\]]*)\s*\]\s*$/,
      param: /^\s*([\w\.\-\_]+)\s*=\s*(.*?)\s*$/,
      comment: /^\s*;.*$/ } ;
  settingsFile = new File($.fileName.replace (/[^\\\/]+$/i , sCfgFile)); // method Ric Quatro
  settingsFile.open("r");

  while(!settingsFile.eof) { 
    thisLine = settingsFile.readln();                  
    if (thisLine === null || thisLine === "") {   // skip blank lines
      continue;
    }
    if (regex.comment.test(thisLine))  {          // skip comment lines
      continue;
    } 
    if (regex.section.test(thisLine))  {          // ignore section lines
      continue;
    }

    match = thisLine.match(regex.param);          // ignore undefined contents

    if (match[1] == sValueName) {
      sFoundValue = match[2];
      notFound = false;
      break;                                      // no need to continue while loop
    }
  }
  settingsFile.close();
  if (notFound) {
    Console ("GetIniValue «" + sValueName + "» not found in configuration file:\n" + sCfgFile);
    return vDefault;
  }
  return sFoundValue;
} // --- end KLD_E.GetIniValue

KLD_E.GetLevel = function (oPgf) { // === Get the level from the ¶ tag name =======================
// Arguments  oPgf           Pargraph from which to get the level by property
// Returns    Title level as number (0 ... 3 or else)
// Called by  KLD_E.BuildTree
var lvl;

 lvl = oPgf.UserString;
 return Number(lvl);
} // end of KLD_E.GetLevel

KLD_E.GetLevelSettings = function () {
// Arguments  sFormats       string version of the format table:
//                           0Format-a 1Format-b 1Format-c 2Format-d 3Format-e
// Comment    sFormats allows easy comparison without loop by means of IndexOf 
// Called by  KLD_E.DocNav
// Calling    KLD_E.GetIniValue
var j, sCfgFile, sFormat, sFormats = "";

  sCfgFile = KLD_E.sExtrasCfg
  for (j = 0; j < 4; j++) {                     // levels 0 … 3
    sFormat = KLD_E.GetIniValue (sCfgFile, j, undefined);
    if (sFormat != undefined) {
      sFormat = sFormat.replace(/;/g, ";"+j);   // there may be mulitple format names
      sFormats = sFormats + j + sFormat + "; "  // ; only for visual aid on inspection
    }
  }
  if (sFormats == "") {
    alert (localize (KLD_E.sDnav_lvlDefs), localize (KLD_E.sDnav_title), true);
    return false;
  }
  KLD_E.sFormats = sFormats;
  return true;
} // end of KLD_E.GetLevelSettings

KLD_E.GetPalSettings = function (oSettings) { // === Get positional setting for the palette =======
// Arguments  oSettings      Settings object
//             .nTop         pixels from top
//             .nLeft        pixels from the left
//             .nTheight     Tree height
//             .nTwidth      Tree width
// Called by  KLD_E.DocNav
// Calling    KLD_E.GetIniValue
var sCfgFile = KLD_E.sExtrasCfg, pos;

  oSettings.nTwidth  = Number(KLD_E.GetIniValue (sCfgFile, "Tree_width", 250));
  oSettings.nTheight = Number(KLD_E.GetIniValue (sCfgFile, "Tree_height", 500));

  oSettings.nTop  = Number(KLD_E.GetIniValue (sCfgFile, "DocNav_top", 10));
  oSettings.nLeft = KLD_E.GetIniValue (sCfgFile, "DocNav_left", "-x");  // + 60 for gap
  if (oSettings.nLeft.substr(0, 1) == "+") {
    pos = Number(oSettings.nLeft.substr(1));
    oSettings.nLeft = app.ScreenWidth + pos;
  } else if (oSettings.nLeft.substr(0, 1) == "-") {
    oSettings.nLeft = app.ScreenX - oSettings.nTwidth - 70; // additional framing around tree
  } else {
    oSettings.nLeft = Number(oSettings.nLeft);
  }
  if (oSettings.nTheight < 250) {
    oSettings.nTheight = app.ScreenHeight - 165;  // additional framing around tree
  }
} // end of KLD_E.GetPalSettings

KLD_E.GetParagraphs = function (oDoc, aPgfs, sFormats) { // Get paragraphs according to table =====
// Arguments  oDoc           The document to be analysed and navigated - the current document
//            aPgfs          Array of Paragraph objects fulfilling the Title criteria
//            sFormats       Table to find the title paragraphs
// Returns    aPgfs
// Called by  KLD_E.BtnCollaps

var oPgf = oDoc.MainFlowInDoc.FirstTextFrameInFlow.FirstPgf;

  while (oPgf.ObjectValid ()) {
    iPos = sFormats.indexOf(oPgf.Name, 0);
    if (iPos > 0) {
      oPgf.UserString = sFormats.substr(iPos-1, 1); // level is just before format name
      aPgfs.push (oPgf);
    }
    oPgf = oPgf.NextPgfInFlow;
  }
  return;
} // end of KLD_E.GetParagraphs

KLD_E.GetText = function (oText, oDoc) { // === Gets the text from the text object. ===============
// Arguments  oText          Any text object, particularly paragraphs
//            oDoc           Document containing oText
// Returns    Text string: text from oText 
// Called by  KLD_E.BuildTree
// Reference  Rick Quatro
var text = "";
  if (oText.constructor.name !== "TextRange") { // Get a list of the strings in the text object or text range.
    var textItems = oText.GetText(Constants.FTI_String);
  } else {
     var textItems = oDoc.GetTextForRange(oText, Constants.FTI_String);
  }
  for (var i = 0; i < textItems.len; i += 1) { // Concatenate the strings.
    text += (textItems[i].sdata);
  }
  return text; // Return the text
} // end of KLD_E.GetText

KLD_E.Navigate = function (oDoc, oTree) { // === Perform the navigation on clik on the tree-item ==
// Arguments  oDoc           The document to be analysed and navigated - the current document
//            oTree          TreeView object to be handled
// Comment    Althoug the selection is cancelled, it still can be recognised as grey shadow (fine so)
// Called by  oTree.onChange in KLD_E.BtnCollaps
// Reference  Rick Quatro: The constant ":2" is from Hypertext marker syntax. 
//            When you generate a TOC and look at the markers, you will see the 2. 
//            If you generate a list of markers or a list of elements, the number is different:
//            The number depends on the type of object that you are referencing.
var unique, docName, item, cmd;
//$.bp(true);
  item = oTree.selection;
//$.writeln ("Navigate " + item.text);
  if (item) {
    navName    = item.navName;
    navUnique  = item.navUnique;
    cmd = "gotoObjectId " + navName + ":2 " + navUnique;
    oDoc.HypertextCommand (cmd);
    oTree.selection = null;
  }
} // end of KLD_E.Navigate

KLD_E.PutIniValue = function (sCfgFile, sValueName, sValue) {
// Arguments  sCfgFile       ini file name (e.g. "
//            sValueName     key that is used in the ini-file (e.g. "FMhome")
//            sValue         value of the key
// Attention  This function does not observe ini sections. Hence the key names must be unique
//            If a key does not already exist, a message is issued and the key is added at the end.
// Comment    File errors reported only to Console - en only
// Returns    false on error
// Attention  Windows ini files are still only CP 1252, whareas maker.ini can be UTF-8 also
// Used in    KLD_E.DefineLevels
var aLines=[], aMatch, j, oRegEx, fSettings, nLines, sFile, notFound = true; 

  sCfgFile = KLD_E.sExtrasCfg;
  oRegEx = {
    section: /^\s*\[\s*([^\]]*)\s*\]\s*$/,
    param: /^\s*([\w\.\-\_]+)\s*=\s*(.*?)\s*$/,   // only this used here
    comment: /^\s*;.*$/ } ;
  sFile = $.fileName.replace (/[^\\\/]+$/i , sCfgFile); // method Ric Quatro
  fSettings = new File(sFile); // method Ric Quatro
  fSettings.open("r");
  if (fSettings.error !== "") {
    Console ("Open read error for file " + sFile);
    return false;
  }
  while(!fSettings.eof) {                         // read entire file into array
    aLines.push(fSettings.readln());
  } 
  fSettings.close();
  
  nLines = aLines.length;

  for (j = 0; j < nLines; j++ ) {                 // Does the key exist? --------------------------
    aMatch = aLines[j].match(oRegEx.param);
    if (aMatch == null) {continue;}
    if (aMatch[1].toUpperCase() == sValueName.toUpperCase()) {
      bNotFound = false;
      aLines[j] = sValueName + " = " + sValue;    // replace this line
      break;
    }
  }
  if (bNotFound) {                                // key does not exsit ---------------------------
    Console ("Value name " + sValueName + " not found in file «" + sCfgFile + "». Item added at the end"); 
    aLines.push (sValueName + " = " + sValue);
  }
  fSettings.open("w");                            // Write modified array -------------------------
  if (fSettings.error !== "") {
    Console ("Open write error for file " + sFile);
    return false;
  }
  nLines = aLines.length;
  for (j = 0; j < nLines; j++ ) {
    fSettings.writeln(aLines[j]);
  }
  fSettings.close();
  return true;
} // end of KLD_E.PutIniValue

KLD_E.SetUpMenu = function (){ // === Create menu entry ==========================================
var menuLocation, oCmd = {};

// ---------------------------------------------- Menu and Command definition for documents -------
  menuLocation  = app.GetNamedMenu("ViewMenu"); 
  oCmd.MenuDnav = menuLocation.DefineAndAddCommand (1,"ETBdocNav",  localize (KLD_E.sDnav_title), "\\!qdn");
  oCmd.MenuDnav.KeyboardShortcutLabel = "ESC q d n";  
  UpdateMenus();
} // --- end KLD_E.SetUpMenu

function Command(cmd) { // ========================================================================
	switch(cmd)  {	                        // Respond to menu command.
		case 1: 
      KLD_E.DocNav ();
      break;
    }
} // --- end Command


KLD_E.SetUpMenu ();                               // comment for debug without menu
//KLD_E.DocNav ();                                // uncomment for debug without menu
